#include <iostream.h>
#include <fstream.h>
#include <queue.h>
#include <list.h>
#include <assert.h>

#include "graph.h"
#include "../Util.h"
#include "../KeyValue.h"

bool node_pair::parseStr(const char *s) {
  vector<JString> fields;
  int num = str_split_char(s, fields, ':');
  if(num != 2) {
    return false;
  }
  v_id = atol(fields[0].cstr());
  u_id = atol(fields[1].cstr());
  return true;
} // End node_pair::parseStr()

JString node_pair::toStr() {
  char buf[1024];
  sprintf(buf, "%ld:%ld", v_id, u_id);
  JString toret(buf);
  return toret;
} // End node_pair::toStr()

void edge::addProperty(const char *key, const char *val) {
  KeyValue_t kv(s.cstr());
  kv.add_kv_pair(key, val);
  s = kv.toStr();
} // End edge::addProperty()

void graph::readGraph(const char *fname) {

  ifstream g_fin(fname);

  for(;;) {
    // read each line
    if(g_fin.eof()) break;
    char line[8192*2];
    g_fin.getline(line, sizeof(line)); // read a line
    if(g_fin.eof()) break;
    if(g_fin.fail()) {
      cerr << "Error reading " << fname << endl;
      cerr << line << endl;
      exit(99);
    }

    // line has format:
    // "<dest-node-id> <node-info-str> <origin-node1> <colour1> <origin-node2> <colour2>..."
    vector<JString> sVec;
    int numFields = split(line, sVec);
    if(numFields <= 0) {
      cerr << "Format of " << fname << " is wrong" << endl;
      exit(97);
    }
    long dest_node = JString_atoi(sVec[0]);
    if(!nodeExists(dest_node)) { addNode(dest_node); }

    int n_ind = getNodeIndex(dest_node);
    assert(n_ind >= 0);
    node_t &n = V[n_ind];
    JString &s = sVec[1];
    assert(s[0] == '[');
    assert(s[s.length()-1] == ']');
    n.s = s.substring(1, s.length()-1);

    assert(((numFields-2) % 2) == 0);
    for(int i = 2; i < numFields; i+=2) {
      edge_t e;
      e.end_pt = JString_atoi(sVec[i]);
      JString &s1 = sVec[i+1];
      assert(s1[0] == '[');
      assert(s1[s1.length()-1] == ']');
      e.s = s1.substring(1, s1.length()-1);
      e.colour = -1;
      e.cost = -1;

      n.nbrs.push_back(e);
    } // End for(i) reading all edges  

  } // End for(;;) reading all lines in 'graphFile'
  g_fin.close();
} // End graph::readGraph()


void graph::writeGraph(const char *fname) {
  ofstream g_fout(fname);
  for(unsigned int v_index = 0; v_index < V.size(); v_index++) {
    vector<edge_t> &nbrs = V[v_index].nbrs;
    char c = (nbrs.size() == 0) ? '\n' : ' ';
    g_fout << V[v_index].id << " [" << V[v_index].s << "]" << c;
    for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
      char c1 = (e_ind == nbrs.size()-1) ? '\n' : ' ';
      g_fout << nbrs[e_ind].end_pt << " [" << nbrs[e_ind].s << "]" << c1;
    }
  }
  g_fout.close();
} // End graph::writeGraph()


bool graph::addNode(long v_id) {
  if(nodeIdHash.containsKey(v_id)) { return false; }
  node_t n;
  n.id = v_id;
  int n_index = V.size();
  V.push_back(n);
  nodeIdHash.put(v_id, n_index);
  return true;
} // End graph::addNode()


bool graph::removeNode(long v_id) {
  int orig_ind = getNodeIndex(v_id);
  if(orig_ind < 0) { return false; } // node not present

  // Remove the node, and all edges
  for(vector<node_t>::iterator n_iter = V.begin(); n_iter != V.end(); n_iter++) {
    if(n_iter->id != v_id) { continue; }
    vector<edge_t> &nbrs = n_iter->nbrs;
    // Remove all edges to v_id in graph
    for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
      if(nbrs[e_ind].end_pt != v_id) {
	// Self loops get removed when the node is removed
	remove_uni_edge(nbrs[e_ind].end_pt, v_id);
      }
    }
    V.erase(n_iter); // remove node itself
    break;
  } // End for(n_iter)

  for(unsigned int i = orig_ind; i < V.size(); i++) {
    // update nodeIdHash
    int ind = getNodeIndex(V[i].id);
    if(ind != (int)(i+1)) {
      fprintf(stderr, "graph::removeNode: ind != i+1 panic!\n");
      exit(70);
    }
    nodeIdHash.put(V[i].id, i); // update index
  }
  nodeIdHash.remove(v_id); // remove the nodeIdHash of the deleted node

  return true;
} // End graph::removeNode()


bool graph::nodeExists(long v_id) {
  if(!nodeIdHash.containsKey(v_id)) { return false; }
  int v_index = nodeIdHash.get(v_id);
  return (v_index >= 0) && ((unsigned int)v_index < V.size());
} // End graph::nodeExists()


int graph::getNodeIndex(long v_id) {
  if(!nodeIdHash.containsKey(v_id)) { return -1; }
  int v_index = nodeIdHash.get(v_id);
  return ((v_index >= 0) && ((unsigned int)v_index < V.size())) ? v_index : -1;
} // End graph::getNodeIndex()


node_t *graph::getNode(long v_id) {
  int ind = getNodeIndex(v_id);
  return (ind == -1) ? NULL : &(V[ind]);
} // End graph::getNode()


bool graph::setNode_s(long v_id, const JString &s) {
  int v_index = getNodeIndex(v_id);
  if(v_index == -1) { return false; }
  V[v_index].s = s;
  return true;
} // End graph::setNode_s()


int graph::getDegree(long v_id) {
  int v_index = getNodeIndex(v_id);
  return (v_index >= 0) ? (int)(V[v_index].nbrs.size()) : -1;
} // End graph::getDegree()


void graph::getPoint(long id, Point *p){
  int v_index = getNodeIndex(id);
  Point p1(V[v_index].s.cstr());
  p->copyFrom(p1);
}//End graph::getPoint()

void graph::getLineVec(long id1, long id2, LineVec *lv){
  Point p1, p2;
  getPoint(id1,&p1);
  getPoint(id2,&p2);
  LineVec lv1(p1,p2);
  lv->copyFrom(&lv1);
}//End graph::getLineVec()

int graph::numEdges() {
  int count = 0;
  for(unsigned int v_index = 0; v_index < V.size(); v_index++) {
    count += V[v_index].nbrs.size();
  }
  return count;
} // End graph::numEdges()


bool graph::add_edge(long v_id, long u_id) {
  return add_uni_edge(v_id, u_id) && add_uni_edge(u_id, v_id);
} // End graph::add_edge()


bool graph::add_uni_edge(long v_id, long u_id) {

  int v_index = getNodeIndex(v_id);
  if(v_index == -1) { return false; }
  vector<edge_t> &nbrs = V[v_index].nbrs;
  for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
    if(nbrs[e_ind].end_pt == u_id) { return false; }
  }

  edge_t new_edge;
  new_edge.end_pt = u_id;
  nbrs.push_back(new_edge);
  return true;

} // End graph::add_uni_edge()


bool graph::edge_exists(long v_id, long u_id) {
  return uni_edge_exists(v_id, u_id) && uni_edge_exists(u_id, v_id);
} // End graph::edge_exists()


bool graph::uni_edge_exists(long v_id, long u_id) {
  int v_index = getNodeIndex(v_id);
  if(v_index == -1) { return false; }
  vector<edge_t> &nbrs = V[v_index].nbrs;
  for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
    if(nbrs[e_ind].end_pt == u_id) { return true; }
  }
  return false;
} // End graph::uni_edge_exists()


bool graph::remove_edge(long v_id, long u_id) {
  return remove_uni_edge(v_id, u_id) && remove_uni_edge(u_id, v_id);
} // End graph::remove_edge()


bool graph::remove_uni_edge(long v_id, long u_id) {

  int v_index = getNodeIndex(v_id);
  if(v_index == -1) { return false; }
  vector<edge_t> &nbrs = V[v_index].nbrs;
  for(vector<edge_t>::iterator iter = nbrs.begin(); iter != nbrs.end(); iter++) {
    if(iter->end_pt == u_id) {
      nbrs.erase(iter);
      return true;
    }
  }

  return false;

} // End graph::remove_uni_edge()


edge_t *graph::getEdge(long v_id, long u_id) {
  int v_index = getNodeIndex(v_id);
  if(v_index == -1) { return NULL; }
  int e_ind = e_ind_of_nbr(v_id, u_id);
  if(e_ind == -1) { return NULL; }
  return &(V[v_index].nbrs[e_ind]);
} // End graph::getEdge()


int graph::reverse_edge(long v_id, int e_ind) {
  int v_index = getNodeIndex(v_id);
  if(v_index == -1) { return -1; }
  long u_id = V[v_index].nbrs[e_ind].end_pt;
  int u_index = getNodeIndex(u_id);
  if(u_index == -1) { return -1; }
  vector<edge_t> &u_nbrs = V[u_index].nbrs;
  for(unsigned int rev_e_ind = 0; rev_e_ind < u_nbrs.size(); rev_e_ind++) {
    if(u_nbrs[rev_e_ind].end_pt == v_id) {
      return rev_e_ind;
    }
  }
  return -1;
} // End graph::reverse_edge()


unsigned int graph::compute_Delta() {
  unsigned int Delta = 0;
  for(unsigned int v_index = 0; v_index < V.size(); v_index++) {
    vector<edge_t> &nbrs = V[v_index].nbrs;
    if(Delta < nbrs.size()) { Delta = nbrs.size(); }
  }
  return Delta;
} // End graph::compute_Delta()


int graph::e_ind_of_nbr(long v_id, long u_id) {
  int v_index = getNodeIndex(v_id);
  if(v_index == -1) { return -1; }
  vector<edge_t> &nbrs = V[v_index].nbrs;
  for(unsigned int i = 0; i < nbrs.size(); i++) {
    if(nbrs[i].end_pt == u_id) { return i; }
  }
  return -1;
} // End graph::e_ind_of_nbr()


void graph::assign_colour(long v_id, int e_ind, int colour) {
  int v_index = getNodeIndex(v_id);
  vector<edge_t> &nbrs = V[v_index].nbrs;
  nbrs[e_ind].colour = colour;

  long u_id = nbrs[e_ind].end_pt;
  int u_index = getNodeIndex(u_id);
  vector<edge_t> &u_nbrs = V[u_index].nbrs;
  int rev_e_ind = reverse_edge(v_id, e_ind);
  u_nbrs[rev_e_ind].colour = colour;
} // End graph::assign_colour()


void graph::get_default_edge_order(vector<node_pair_t> &eVec) {
  while(eVec.size() > 0) { eVec.pop_back(); }

  for(unsigned int v_index = 0; v_index < V.size(); v_index++) {
    vector<edge_t> &nbrs = V[v_index].nbrs;
    for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
      // Each edge appears only once, with v_id < u_id
      if(V[v_index].id >= nbrs[e_ind].end_pt) { continue; }
      node_pair_t np;
      np.v_id = V[v_index].id; np.u_id = nbrs[e_ind].end_pt;
      eVec.push_back(np);
    }
  }
} // End graph::get_default_edge_order()


void graph::bfs_order(long start_v_id, vector<node_pair_t> &eVec) {

  while(eVec.size() > 0) { eVec.pop_back(); }

  // Breadth First Search, starting from node start_v_id

  // Node depth in the BFS tree -- 0 for start_v_id, -1 if unseen so far
  vector<int> node_depth;

  // Init: designate start_v_id to be seen and insert it in queue
  for(unsigned i = 0; i < V.size(); i++) {
    int depth = (V[i].id == start_v_id) ? 0 : -1;
    node_depth.push_back(depth);
  }
  queue<long, list<long> > bfsQ;
  bfsQ.push(start_v_id);

  while(!bfsQ.empty()) {

    // Pop the front element of the queue, expand it, add the
    // neighbours that are not already seen
    long v_id = bfsQ.front();
    bfsQ.pop();

    int v_index = getNodeIndex(v_id);
    vector<edge_t> &nbrs = V[v_index].nbrs;
    int v_depth = node_depth[v_index];

    // Sort edges in order of decreasing degree
    vector<int> sort_order;
    for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
      sort_order.push_back(e_ind);
    }
    for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
      for(unsigned int e_ind1 = e_ind+1; e_ind1 < nbrs.size(); e_ind1++) {
	long u_id = nbrs[sort_order[e_ind]].end_pt;
	long u_id1 = nbrs[sort_order[e_ind1]].end_pt;
	int deg = getDegree(u_id);
	int deg1 = getDegree(u_id1);
	if(deg1 > deg) {
	  int temp = sort_order[e_ind];
	  sort_order[e_ind] = sort_order[e_ind1];
	  sort_order[e_ind1] = temp;
	}
      }
    }

    for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
      long u_id = nbrs[sort_order[e_ind]].end_pt;
      int u_index = getNodeIndex(u_id);
      int u_depth = node_depth[u_index];

      if(u_depth == -1) {
	// So far unseen node, edge needs to be inserted into eVec
	node_pair_t np;
	np.v_id = v_id; np.u_id = u_id;
	eVec.push_back(np);
	//fprintf(stderr, "bfs:new_node[v_id=%ld, u_id=%ld]\n", v_id, u_id);
	// Also, need to add the new node to bfsQ
	bfsQ.push(u_id);
	// Update node_depth for u_id
	node_depth[u_index] = v_depth+1;

      } else if(u_depth == v_depth) {
	// Insert this edge deterministically
	if(v_id < u_id) {
	  node_pair_t np;
	  np.v_id = v_id; np.u_id = u_id;
	  eVec.push_back(np);
	  //fprintf(stderr, "bfs:new_edge_same_depth[v_id=%ld, u_id=%ld]\n", v_id, u_id);
	}

      } else if(u_depth == (v_depth+1)) {
	node_pair_t np;
	np.v_id = v_id; np.u_id = u_id;
	eVec.push_back(np);
	//fprintf(stderr, "bfs:new_edge_next_depth[v_id=%ld, u_id=%ld]\n", v_id, u_id);
      } else {
	// No need to insert this edge in eVec since reverse edge
	// would have been considered
	if(u_depth != (v_depth-1)) {
	  fprintf(stderr, "bfs:err[v_id=%ld, u_id=%ld]\n", v_id, u_id);
	  fprintf(stderr, "v_depth=%d, u_depth=%d\n", v_depth, u_depth);
	  exit(12);
	}
      }

    } // End for(e_ind)

  } // End while(queue is not empty)

} // End graph::bfs_order()

void graph::shortest_path_tree(long start_v_id, vector<treeNode_t> &nVec) {

  while(nVec.size() > 0) { nVec.pop_back(); }

  // Breadth First Search, starting from node start_v_id

  // Init: designate start_v_id to be seen and insert it in queue
  for(unsigned i = 0; i < V.size(); i++) {
    int depth = (V[i].id == start_v_id) ? 0 : -1;
    treeNode_t tnode;
    tnode.depth = depth;
    tnode.id = V[i].id;
    tnode.ll = start_v_id;
    nVec.push_back(tnode);
  }
  queue<long, list<long> > bfsQ;
  bfsQ.push(start_v_id);

  int traversal_index = 0;

  while(!bfsQ.empty()) {
    // Pop the front element of the queue, expand it, add the
    // neighbours that are not already seen
    long v_id = bfsQ.front();
    bfsQ.pop();
    
    int v_index = getNodeIndex(v_id);
    vector<edge_t> &nbrs = V[v_index].nbrs;
    if(nVec[v_index].id != v_id){
      fprintf(stderr,"shortest path tree: index don't match\n");
      exit(202);
    };    
    int v_depth = nVec[v_index].depth;

    nVec[v_index].traversal_index = traversal_index;
    traversal_index++;

    // Sort edges in order of decreasing degree
    vector<int> sort_order;
    for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
      sort_order.push_back(e_ind);
    }
    for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
      for(unsigned int e_ind1 = e_ind+1; e_ind1 < nbrs.size(); e_ind1++) {
	long u_id = nbrs[sort_order[e_ind]].end_pt;
	long u_id1 = nbrs[sort_order[e_ind1]].end_pt;
	int deg = getDegree(u_id);
	int deg1 = getDegree(u_id1);
	if(deg1 > deg) {
	  int temp = sort_order[e_ind];
	  sort_order[e_ind] = sort_order[e_ind1];
	  sort_order[e_ind1] = temp;
	}
      }
    }//sort

    for(unsigned int e_ind = 0; e_ind < nbrs.size(); e_ind++) {
      long u_id = nbrs[sort_order[e_ind]].end_pt;
      int u_index = getNodeIndex(u_id);
      if(nVec[u_index].id != u_id){
	fprintf(stderr,"shortest path tree: index don't match\n");
	exit(202);
      };
      int u_depth = nVec[u_index].depth;
      
      if(u_depth == -1) {
	// So far unseen node, update child/parent info
	nVec[u_index].parent = v_id;
	vector<long> &cpr = nVec[v_index].children;
	long child = u_id;
	cpr.push_back(child);
	
	//fprintf(stderr, "bfs:new_node[v_id=%ld, u_id=%ld]\n", v_id, u_id);
	// Also, need to add the new node to bfsQ
	bfsQ.push(u_id);
	// Update node_depth for u_id
	nVec[u_index].depth = v_depth+1;
      } 
    } // End for(e_ind)
  } // End while(queue is not empty)
  // expand children list
  queue<long, list<long> > childQ;
  int status[nVec.size()];
  for (unsigned i =0; i< nVec.size();i++) {status[i]=0;}
  for(unsigned i = 0; i < nVec.size(); i++) {
    while(childQ.size() > 0) { childQ.pop(); }
    for(unsigned j=0; j < nVec[i].children.size(); j++){
      childQ.push(nVec[i].children[j]);
    }
    while(!childQ.empty()) {
      long v_id = childQ.front(); 
      childQ.pop();
      int v_index = getNodeIndex(v_id);
      for(unsigned j=0; j < nVec[v_index].children.size();j++){
	if(status[v_index]==0) childQ.push(nVec[v_index].children[j]);
	vector<long> &cpr = nVec[i].children;
	long child = nVec[v_index].children[j];
	cpr.push_back(child);
      }
    }
    status[i] = 1;
  }//for i
} // End graph::shortest_path_tree()

bool graph::check_bipartite() {
  vector<node_pair_t> nprs;
  int *colour = new int[V.size()];
  //uncolour all nodes first
  for(unsigned int i = 0; i < V.size(); i++){
    colour[i]=-1;
  }//for(i<V.size())
  colour[0]=0;
  //get the BFS order and then check if every edge has opposite coloured
  //nodes
  bfs_order(V[0].id,nprs);
  for(unsigned int i = 0; i < nprs.size(); i++){
    int v_index = getNodeIndex(nprs[i].v_id);
    int u_index = getNodeIndex(nprs[i].u_id);
    if(colour[v_index]==-1){
      if(colour[u_index]==-1){
	fprintf(stderr,"both nodes are not coloured: error");
	exit(10);
      }//if both uncoloured
      colour[v_index]=1-colour[u_index];
    }//if v_index
    
    if(colour[v_index] > -1){
      if(colour[u_index]==-1){
	colour[u_index]=1-colour[v_index];
      }//if u_index uncoloured
      if(colour[v_index]==colour[u_index]){
	delete [] colour; 
	return false;
      }//if both coloured
    }//if v_index
  }//for(i< nprs.size)
  delete [] colour; 
  return true;
}//End graph::check_bipartite

void graph::copy_graph(class graph *g1) {
  while(V.size() > 0) { V.pop_back(); }
  for(unsigned int i = 0; i < (g1->V).size(); i++) {
    vector<edge_t> &nbrs = (g1->V)[i].nbrs;
    node_t n;
    for(unsigned int i = 0; i < nbrs.size(); i++) {
      n.nbrs.push_back(nbrs[i]);
    }
    V.push_back(n);
  }
} // End graph::copy_graph()


void graph::copy_colouring(class graph *g1) {
  assert(V.size() == (g1->V).size());
  for(unsigned int i = 0; i < (g1->V).size(); i++) {
    vector<edge_t> &g1_nbrs = (g1->V)[i].nbrs;
    vector<edge_t> &nbrs = V[i].nbrs;
    assert(g1_nbrs.size() == nbrs.size());
    for(unsigned int j = 0; j < g1_nbrs.size(); j++) {
      nbrs[j].colour = g1_nbrs[j].colour;
    }
  }
} // End graph::copy_colouring()


void graph::removeAllEdges() {
  vector<node_pair_t> nprs;
  get_default_edge_order(nprs);
  for(unsigned i = 0; i < nprs.size(); i++) {
    remove_edge(nprs[i].v_id, nprs[i].u_id);
  }
} // End graph::removeAllEdges()


