#ifndef GRAPH_H
#define GRAPH_H

#include <vector.h>

#include <stdio.h>

#include "../JHashtable.h"
#include "../JString.h"
#include "../Util.h"
#include "../KeyValue.h"
#include "../Point.h"
#include "../LineVec.h"

struct edge {
  long end_pt;
  JString s; // Generic string to store info
  int index; // for use in external program to maintain edge specific info
  double cost; // for cost related algos
  int colour; // for edge colouring related algos
  edge() { end_pt = -1;  index = -1; }
  void addProperty(const char *key, const char *val);
};
typedef struct edge edge_t;

struct node {
  long id; // >= 0
  JString s; // Generic string to store info
  int index; // for use in external program to maintain node specific info
  int colour; // for node colouring related algos
  node() { colour = -1; index = -1; id = -1;}
  vector<edge_t> nbrs;
};
typedef struct node node_t;

struct node_pair {
  long v_id, u_id;
  node_pair() { v_id = u_id = -1; }
  node_pair(const char *s) { v_id = u_id = -1; parseStr(s); }
  /** Parse the given v_id:u_id string and assign v_id and u_id
      appropriately.  If no ':' is present, or more than one ':' is
      present, 'false' is returned.  Any other format error produces
      unpredictable result (may return 'true').  Returns 'true' on
      success. */
  bool parseStr(const char *s);
  /** Return v_id:u_id string */
  JString toStr();
};
typedef struct node_pair node_pair_t;

struct treeNode {
  long id; // >= 0
  JString s; // Generic string to store info
  int index; // for use in external program to maintain node specific info
  long ll; // landline
  long parent;
  vector<long> children;
  int depth; // wrt landline
  int traversal_index; // the index in BFS traversal
  treeNode() { depth = -1; index = -1; id = -1; parent = -1; traversal_index = -1; }
};
typedef struct treeNode treeNode_t;

struct graph {
  vector<node_t> V; // Array of vertices
  JHashtable<unsigned long, int> nodeIdHash;
  graph() {
    nodeIdHash.setHashFunction(ulongHash); // Init the hashtable
  }

  KeyValue_t vizingOpts;

  /*---- I/O routines ----*/
  // Read the given graph file, return success
  void readGraph(const char *fname);
  // Write graph to given file, return success
  void writeGraph(const char *fname);

  /*---- Graph manipulation routines ----*/

  // Add a node, return success
  bool addNode(long id);
  // Delete node, return success
  bool removeNode(long id);
  // Does node exist?
  bool nodeExists(long id);
  // Get index of node, into vector 'V', return -1 on failure
  int getNodeIndex(long id);
  // Set node's 's', return success
  bool setNode_s(long id, const JString &s);
  // Get the node corresponding to the given structure, NULL on failure
  node_t *getNode(long id);
  // Return the number of nodes in the graph
  int numNodes() { return V.size(); };

  // Set edge's 's'. return success
  bool setEdge_s(long id1, long id2, const JString &s);

  // Get the degree of given node, -1 on failure
  int getDegree(long id);

  // Return the coordinates of the given node in "p"
  void getPoint(long id, Point *p);

  // Get the line vector corresponding to the two nodes
  void getLineVec(long id1, long id2, LineVec *lv);
	       

  // Return the number of edges in the graph (an undirected edge is
  // counted twice)
  int numEdges();

  // Add bidirectional edge, return failure if edge already present in either direction
  // NOTE: resulting graph undefined if edge in one direction present
  bool add_edge(long v_id, long u_id);
  // Add unidirectional edge if possible, return failure if already present
  bool add_uni_edge(long v_id, long u_id);

  // Does given edge exist?  Check both directions
  bool edge_exists(long v_id, long u_id);
  // Does given unidirectional edge exist?
  bool uni_edge_exists(long v_id, long u_id);

  // Remove given edge, return failure edge absent in either direction
  // NOTE: resulting graph undefined if edge in one direction present
  bool remove_edge(long v_id, long u_id);
  // Remove unidirectional edge, return failure if edge absent
  bool remove_uni_edge(long v_id, long u_id);

  /*---- Some utility routines ----*/

  // Return the edge structure corresponding to the given nodes, NULL
  // on failure
  edge_t *getEdge(long v_id, long u_id);

  // Return the reverse edge of the given edge at the given node
  // Return -1 on failure
  int reverse_edge(long v_id, int e_ind);

  // compute the max degree in the graph
  unsigned int compute_Delta();

  // Get the e_ind corresponding to the given neighbour, -1 on failure
  int e_ind_of_nbr(long v_id, long u_id);

  // Assign colour to given edge and the reverse edge
  void assign_colour(long v_id, int e_ind, int colour);

  // Compute default edge ordering
  // Note: in the returned vector of node_pairs, v_id < u_id in each
  // element
  void get_default_edge_order(vector<node_pair_t> &eVec);

  // Return Bread First Search order of edges, starting from node
  // start_v_id -- return in eVec
  void bfs_order(long start_v_id, vector<node_pair_t> &eVec);

  // Calculates parent/child information (returned in nVec) of a
  // shortest path tree rooted at start_v_id
  void shortest_path_tree(long start_v_id, vector<treeNode_t> &nVec);

  // Is the graph bipartite?
  bool check_bipartite();

  // Copy graph structure from given graph
  void copy_graph(class graph *g1);
  // Copy colouring from given graph (assume its a copy of 'this')
  void copy_colouring(class graph *g1);

  /** Get an empty graph with just isolated vertices, by removing all
      edges from this graph */
  void removeAllEdges();

};
typedef struct graph graph_t;

#endif
