#ifndef WP2P_LL_H
#define WP2P_LL_H

#include "wp2p-mac.h"
#include <ll.h>
#include <packet.h>
#include <tcp.h>
#include <marshall.h>

/** How many packets to store on the way up or down? */
/** This has a maximum limit of 200 pkt **/

#define MAX_PKTS 700

/** This structure stores the ack seqno and the ack window in the
    received packets (these acks are obviously for packets that were
    sent) */
struct recv_info {
  uint16_t ackseq;
  uint32_t ackwin;
};
typedef struct recv_info recv_info_t;

/** The sequence number is 16 bits long, with wrap around.  These
    define the successor and predecessor relationships in the wrap
    around scenario. */
#define succ16(x) (((x) == 0xffff) ? 0 : ((x)+1))
#define pred16(x) (((x) == 0) ? 0xffff : ((x)-1))

inline uint16_t get_seqno(Packet *p) {
  hdr_mac2P *llh = HDR_MAC2P(p);
  // convert u_char to seqno;
  uint16_t *p3 = (uint16_t*)(llh->dh_bssid);
  uint16_t dataseq = ntohs(*p3);
  return dataseq;
} // End get_seqno()

typedef Packet *Packet_p;
typedef Handler *Handler_p;

/** Packets need to be stored on the way up or down the stack.  This
    structure defines the relevant information. */
struct pkt_info {
  /** Pointer to the packet*/
  Packet *p;

  /** The Handler may not be required, I do not know -- I just added
      it anyway. */
  Handler *h;

  /* The packet sequence number */
  uint16_t seqno;

  /** Number of times packet has been transmitted (used in downQ) */
  int num_tx;
  /** Number of phases for which pkt has been waiting around in buf
      (used in upQ) */
  int num_phases_in_buf;

  /* Does this structure contain a valid packet information? */
  bool free;

  /** Constructor, initialization. */
  pkt_info() { free = true; num_tx = 0; num_phases_in_buf = 0; }
};
typedef struct pkt_info pkt_info_t;


/** How many times to retry when ack is not received?  Note: this is
    the number of retries, not the number of tries.  Number of tries
    will be one more than this value. */
#define MAX_RETRY 4

/** The structure to maintain the buffer of packets.  I'm using the
    same structure for both directions (although there will be two
    different buffers -- see upQ and downQ below). */
class UpDownBuf {
public:
  /** The buffer of packets stored */
  pkt_info_t buf[MAX_PKTS];
  
  int blim_;              /* maximum allowed pkts in queue */
  
  /** Constructor, initialization */
  UpDownBuf();

  /** Add the given packet with the given sequence number to the
      buffer */
  bool addPkt(uint16_t dataseq, Packet *p, Handler *h);

  /*---- These are routines used by upQ ----*/
  /** What is the minimum sequence number?  The minimum sequence
      number is returned in '*min_seq_p'.  The return value is the
      index into the buffer with the min entry, -1 if buf is empty.
      Return the num_phases_in_buf of the min seqno, in
      '*num_phases_in_buf_p', if num_phases_in_buf_p is not NULL.
      Return the num_tx of the min_seqno, in '*num_tx_p', of num_tx_p
      is not NULL. */
  int minSeqNo(uint16_t *min_seq_p, int *num_phases_in_buf_p, int *num_tx_p);
  /** Remove the minimum sequence number and return it in '*pp'.  The
      handler is returned in '*hp'.  Return success/failure. */
  bool removeMin(Packet_p *pp, Handler_p *hp);
  /** Return the index into 'buf' of the given sequence number.
      Return -1 on failure. */
  int indexOf(uint16_t seq);
  /** Is the given sequence number present in 'buf' ? */
  bool has_seqno(uint16_t seq);
  /** Increment the phase count (by one) for all packets in the upQ */
  void incr_phase_count();

  /*---- These are routines used by downQ ----*/
  /** Is 'buf' full? */
  bool full();
  /** Remove the given sequence from 'buf'.  Return the packet in
      *pp, and the handler in *hp.  Return success/failure. */
  bool remove(uint16_t seq, Packet_p *pp, Handler_p *hp);

  /** Sort the buffer in order of sequence numbers.  Also, "empty"
      entries in 'buf' will be pushed to the back of the buffer.  This
      sorting is done so that 'getData' can then simply get the first
      set of packets from 'buf' (Note: this relates to a particular
      strategy for which packets LLC will send in each PHASE) */
  void sort_buf();
  /** Get the data to be sent to MAC layer for sending DOWN.  The
      packets will all be returned in P[], the handlers in H[], and
      sequence numbers in sendseq[].  Note that 'P'is an array of
      'Packet_p', and not arrays of 'Packet'.  'num_pkts' is an 'IN'
      parameter that defines the number of packets to be returned in
      this call of 'getData'.  Return value of this function is the
      number of 'valid' (as opposed to dummy/filler) packets returned
      in the 'P', 'H', and 'sendseq' arrays.  The dummy/filler packets
      will be marked with the 'filler' field set to 1 (see
      modification to struct hdr_cmn in "~ns/common/packet.h").  Their
      size will be set to the given dummy_size parameter. */
  int getData(uint16_t num_pkts, Packet_p *P, Handler_p *H,
	      uint16_t *sendseq, int dummy_size);

  bool less16(uint16_t x, uint16_t y);


}; // End class UpDownBuf

/** This is the class that implements the actual LLC layer */
class WP2P_LL : public LL {
public:
  /** Constructor, initialization */
  WP2P_LL();

  /** TCL command processing */
  int command(int argc, const char*const* argv);

  /** Receive a packet (in UP/DOWN direction) */
  virtual void recv(Packet* p, Handler* h);
  /** Send the packet up */
  virtual void sendUp(Packet* p, Handler *h);

  /** Process the received ack -- will affect the downQ */
  virtual void processRecdAck(uint16_t ackseq, uint32_t ackwin);
  /** Process the received data -- will affect the upQ */
  virtual void processRecdData(uint16_t dataseq, Packet *p, Handler *h);
  /** Process the data to be sent -- will take packets from downQ */
  virtual void processSendData();
  /** Process the ack to be sent -- will make use information from the upQ */
  virtual void processSendAck(uint16_t *send_ackseq, uint32_t *send_ackwin);
  /** Fill the appropriate headers, and send the given packet DOWN the
      stack. */
  virtual void sendDown(uint16_t sendseq, Packet *p, Handler *h,
			uint16_t send_ackseq, uint32_t send_ackwin);

  /** Sugar routines  */
  inline int my_node_id() {
    return my_node_addr_;
  }

  inline void set_my_node_id(int id) {
    my_node_addr_ = id;
  }

  inline int nbr_node_id() {
    return nbr_node_addr_;
  }

  inline void set_nbr_node_id(int id) {
    nbr_node_addr_ = id;
  }

  inline int get_pkt_size() {
    return dummyPktSize_;
  }
  
protected:
  WP2P_MAC *mac_;
  /** My MAC address */
  int my_mac_addr_;
  /** My neighbour's MAC address */
  int nbr_mac_addr_;
  /** Node addresses */
  int nbr_node_addr_,my_node_addr_;
  

  /** Sequence number sent DOWN so far.  Range: 0 to 2^16-1 (65535),
      with wrap around. */
  uint16_t seqno_;
  /** Seq no sent UP so far.  Range: 0 to 2^16-1 (65535), with wrap
      around. */
  uint16_t left_win_;

  /** Buffer for packets in UP direction (to attempt in-order delivery) */
  UpDownBuf upQ;
  /** Buffer for packets in DOWN direction (to attempt retransmission) */
  UpDownBuf downQ;

  /** Is higher layer unblocked? */
  bool higher_layer_unblocked;
  
  /** Dummy packet size */
  int dummyPktSize_;

  /** As pkts are being received from the other direction, each will
      have an ack.  Store that ack information here.  'recvseq' gives
      the number of such acks received since the start of the current
      receive mode in the PHASE.  (So far as I see, this is not used
      anywhere, but I have let it stay as of now). */
  recv_info_t R_info[MAX_PKTS_IN_PHASE];
  int recvseq;
  //static FILE *fp; // traces the packet flow
  //static int link_id; //for tracing
  //int my_link_id; // for tracing
}; // End class WP2P_LL

//int WP2P_LL::link_id = 0;
//FILE *WP2P_LL::fp = fopen("output/link.tr","w");

#endif
